import logging
from typing import Any, Callable

import eth_retry
from brownie import chain, web3
from requests import Session
from requests.adapters import HTTPAdapter
from web3 import HTTPProvider, Web3
from web3.middleware import filter

from y.networks import Network
from y.utils.cache import memory

logger = logging.getLogger(__name__)

provider_specific_batch_sizes = {
    "moralis":  2_000,
    "pokt":     2_000,
    "tenderly": 2_000,
    "ankr":     2_000,
}

chain_specific_max_batch_sizes = {
    Network.Mainnet:    10_000,     # 1.58 days
    Network.Gnosis:     20_000,     # 1.15 days
    Network.Fantom:     100_000,    # 1.03 days
    Network.Arbitrum:   20_000,     # 0.34 days
    Network.Optimism:   800_000,    # 10.02 days
}

fallback_batch_size = 10_000

def _get_batch_size() -> int:
    for provider, size in provider_specific_batch_sizes.items():
        if provider in web3.provider.endpoint_uri:
            return size
    return chain_specific_max_batch_sizes.get(chain.id, fallback_batch_size)

BATCH_SIZE = _get_batch_size()


def should_cache(method: str, params: Any) -> bool:
    if method == "eth_getCode" and params[1] == "latest":
        return True
    return False


def cache_middleware(make_request: Callable, web3: Web3) -> Callable:
    @eth_retry.auto_retry
    def middleware(method: str, params: Any) -> Any:
        logger.debug("%s %s", method, params)
        if should_cache(method, params):
            return memory.cache(make_request)(method, params)
        return make_request(method, params)
    return middleware


def setup_middleware() -> None:
    # patch web3 provider with more connections and higher timeout
    if web3.provider:
        try:
            assert web3.provider.endpoint_uri.startswith("http"), "only http and https providers are supported"
            adapter = HTTPAdapter(pool_connections=100, pool_maxsize=100)
            session = Session()
            session.mount("http://", adapter)
            session.mount("https://", adapter)
            web3.provider = HTTPProvider(web3.provider.endpoint_uri, {"timeout": 600}, session)
        except AttributeError as e:
            if "'IPCProvider' object has no attribute 'endpoint_uri'" in str(e):
                pass 
            else:
                raise

    # patch and inject local filter middleware
    filter.MAX_BLOCK_REQUEST = BATCH_SIZE
    web3.middleware_onion.add(filter.local_filter_middleware)
    web3.middleware_onion.add(cache_middleware)

def ensure_middleware() -> None:
    setup_middleware()
    




import asyncio
import logging
from typing import List, Optional

import a_sync
from multicall import Call

from y.classes.common import ERC20, WeiBalance
from y.contracts import has_methods
from y.datatypes import AnyAddressType, Block, UsdPrice

logger = logging.getLogger(__name__)


@a_sync.a_sync(default='sync', cache_type='memory')
async def is_token_set(token: AnyAddressType) -> bool:
    return any(await asyncio.gather(
        has_methods(token, ("getComponents()(address[])", "naturalUnit()(uint)"), sync=False),
        has_methods(token, ("getComponents()(address[])", "getModules()(address[])", "getPositions()(address[])"), sync=False),
    ))

@a_sync.a_sync(default='sync')
async def get_price(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    return await TokenSet(token, asynchronous=True).get_price(block=block)


class TokenSet(ERC20):
    @a_sync.a_sync(ram_cache_maxsize=100)
    async def components(self, block: Optional[Block] = None) -> List[ERC20]:
        components = await self.contract.getComponents.coroutine(block_identifier = block)
        return [ERC20(component, asynchronous=self.asynchronous) for component in components]
    
    async def balances(self, block: Optional[Block] = None) -> List[WeiBalance]:
        if hasattr(self.contract, 'getUnits'):
            balances = await self.contract.getUnits.coroutine(block_identifier = block)
        elif hasattr(self.contract, 'getTotalComponentRealUnits'):
            balances = await asyncio.gather(*[
                Call(self.address, ["getTotalComponentRealUnits(address)(uint)", component.address], block_id=block).coroutine()
                for component in await self.components(block, sync=False)
            ])
        return [WeiBalance(balance, component, block=block) for component, balance in zip(await self.components(block=block, sync=False), balances)]
    
    async def get_price(self, block: Optional[Block] = None) -> UsdPrice:
        total_supply = await self.total_supply_readable(block=block, sync=False)
        if total_supply == 0:
            return UsdPrice(0)
        return UsdPrice(sum(await asyncio.gather(*[balance.__value_usd__(sync=False) for balance in await self.balances(block=block, sync=False)])))




import functools
import os
from typing import Callable, TypeVar, Union
import asyncio

from typing_extensions import ParamSpec

DEBUG = os.environ.get('YPRICEMAGIC_DEBUG', False)

P = ParamSpec('P')
T = TypeVar('T')

def continue_on_revert(func: Callable[P, T]) -> Callable[P, T]:
    '''
    Decorates a call-making function. If the call reverts, returns `None`. Raises any other exceptions.
    '''
    from y.exceptions import continue_if_call_reverted

    if asyncio.iscoroutinefunction(func):
        @functools.wraps(func)
        async def continue_on_revert_wrap(*args: P.args, **kwargs: P.kwargs) -> Union[T, None]:
            try:
                return await func(*args,**kwargs)
            except Exception as e:
                continue_if_call_reverted(e)
    elif callable(func):
        @functools.wraps(func)
        def continue_on_revert_wrap(*args: P.args, **kwargs: P.kwargs) -> Union[T, None]:
            try:
                return func(*args,**kwargs)
            except Exception as e:
                continue_if_call_reverted(e)
    else:
        raise NotImplementedError(f"Unable to decorate {func}")
    return continue_on_revert_wrap



import pytest
from brownie import ZERO_ADDRESS, chain
from multicall import Call

from tests.fixtures import blocks_for_contract
from tests.test_constants import STABLECOINS
from y.classes.common import ERC20
from y.constants import WRAPPED_GAS_COIN, wbtc
from y.contracts import Contract
from y.exceptions import NoProxyImplementation, call_reverted
from y.networks import Network
from y.utils.dank_mids import dank_w3

TOKENS = list(STABLECOINS) + [WRAPPED_GAS_COIN, wbtc.address]

if chain.id == Network.Mainnet:
    # MKR symbol and name methods return bytes, we want to test that our code returns strings
    MKR = "0x9f8F72aA9304c8B593d555F12eF6589cC3A579A2"
    TOKENS.append(MKR)

OLD_SUSD = "0x57Ab1E02fEE23774580C119740129eAC7081e9D3"

TOKENS_BY_BLOCK = [(token, block) for token in TOKENS for block in blocks_for_contract(token)]

@pytest.mark.parametrize('token',TOKENS)
def test_erc20_sync(token):
    token = ERC20(token)

    if token.address == OLD_SUSD:
        pytest.skip('Not applicable to deprecated sUSD.')

    block = chain.height
    assert isinstance(token.contract, Contract), f'Cannot fetch contract for token {token}'
    assert isinstance(token.build_name, str), f'Cannot fetch build name for token {token}'
    assert isinstance(token.symbol, str), f'Cannot fetch symbol for token {token}'
    assert isinstance(token.name, str), f'Cannot fetch name for token {token}'
    assert 10 ** token.decimals == token.scale, f'Incorrect scale fetched for token {token}'
    assert token.total_supply(block) / token.scale == token.total_supply_readable(block), f'Incorrect total supply readable for token {token}'
    assert token.price(), f'Cannot fetch price for token {token}'

@pytest.mark.parametrize('token',TOKENS)
@pytest.mark.asyncio_cooperative
async def test_erc20_async(token):
    token = ERC20(token, asynchronous=True)

    if token.address == OLD_SUSD:
        pytest.skip('Not applicable to deprecated sUSD.')

    block = await dank_w3.eth.block_number
    assert isinstance(token.contract, Contract), f'Cannot fetch contract for token {token}'
    assert isinstance(await token.build_name, str), f'Cannot fetch build name for token {token}'
    assert isinstance(await token.symbol, str), f'Cannot fetch symbol for token {token}'
    assert isinstance(await token.name, str), f'Cannot fetch name for token {token}'
    assert 10 ** await token.decimals == await token.scale, f'Incorrect scale fetched for token {token}'
    assert await token.total_supply(block) / await token.scale == await token.total_supply_readable(block), f'Incorrect total supply readable for token {token}'
    assert await token.price(), f'Cannot fetch price for token {token}'

@pytest.mark.parametrize('token,block',TOKENS_BY_BLOCK)
@pytest.mark.asyncio_cooperative
async def test_erc20_at_block(token, block):
    token = ERC20(token, asynchronous=True)

    if token.address == OLD_SUSD and block >= 13222927:
        pytest.skip('Not applicable to the old sUSD after migration block.')

    # NOTE Some proxy tokens would fail tests in early days because no implementation is specified.
    try:
        if await Call(token.address, 'implementation()(address)', [['imp',None]], block_id = block).coroutine() == ZERO_ADDRESS:
            pytest.skip(f"Not applicable to proxy contracts with implementation not set.")
    except Exception as e:
        if not call_reverted(e):
            raise
    
    # NOTE We've validated token is not problematic proxy, proceed with test.
    try:
        # NOTE also tests ERC20._decimals
        assert await token.total_supply(block) / await token._scale(block) == await token.total_supply_readable(block), f'Incorrect total supply readable for token {token}'
    except NoProxyImplementation:
        pass
    
    assert await token.price(block), f'Cannot fetch price for token {token}'




import logging
from typing import List, Union

from brownie import chain
from lazy_logging import LazyLoggerFactory

from y.networks import Network

yLazyLogger = LazyLoggerFactory("YPRICEMAGIC")

logger = logging.getLogger(__name__)

NETWORK_DESCRIPTOR_FOR_ISSUE_REQ =f'name ({Network.name()})' if Network.name() else f'chainid ({chain.id})'

def gh_issue_request(issue_request_details: Union[str, List[str]], _logger = None) -> None:

    if _logger is None: _logger = logger

    if type(issue_request_details) == str:
        _logger.warning(issue_request_details)

    elif type(issue_request_details) == list:
        for message in issue_request_details:
            _logger.warning(message)

    _logger.warning('Please create an issue and/or create a PR at https://github.com/BobTheBuidler/ypricemagic')
    _logger.warning(f'In your issue, please include the network {NETWORK_DESCRIPTOR_FOR_ISSUE_REQ} and the detail shown above.')
    _logger.warning('and I will add it soon :). This will not prevent ypricemagic from fetching price for this asset.')
